@@ -5,6 +5,8 @@ from django.forms.models import modelformset_factory, inlineformset_factory
 from django.views.generic.detail import SingleObjectMixin, SingleObjectTemplateResponseMixin
 from django.views.generic.list import MultipleObjectMixin, MultipleObjectTemplateResponseMixin
 from django.forms.models import BaseInlineFormSet
+from django.utils.functional import curry
+from .compat import ContextMixin
 
 
 class BaseFormSetMixin(object):
@@ -22,7 +24,7 @@ class BaseFormSetMixin(object):
     can_delete = False
 
     def construct_formset(self):
-        return self.get_formset()(initial=self.get_initial(), **self.get_formset_kwargs())
+        return self.get_formset()(**self.get_formset_kwargs())
 
     def get_initial(self):
         return self.initial
@@ -30,14 +32,30 @@ class BaseFormSetMixin(object):
     def get_formset_class(self):
         return self.formset_class
 
+    def get_extra_form_kwargs(self):
+        """
+        This enables us to use some trickery to pass additional arguments to the form
+        constructor
+        """
+        return {}
+
     def get_form_class(self):
         return self.form_class
 
     def get_formset(self):
-        return formset_factory(self.get_form_class(), **self.get_factory_kwargs())
+        formset_class = formset_factory(self.get_form_class(), **self.get_factory_kwargs())
+        extra_form_kwargs = self.get_extra_form_kwargs()
+
+        # Hack to let as pass additional kwargs to each forms constructor. Be aware that this
+        # doesn't let us provide *different* arguments for each form
+        if extra_form_kwargs:
+            formset_class.form = staticmethod(curry(formset_class.form, **extra_form_kwargs))
+        return formset_class
 
     def get_formset_kwargs(self):
-        kwargs = {}
+        kwargs = {
+            'initial': self.get_initial(),
+        }
         if self.request.method in ('POST', 'PUT'):
             kwargs.update({
                 'data': self.request.POST,
@@ -59,10 +77,7 @@ class BaseFormSetMixin(object):
         return kwargs
 
 
-class FormSetMixin(BaseFormSetMixin):
-    def get_context_data(self, **kwargs):
-        return kwargs
-
+class FormSetMixin(BaseFormSetMixin, ContextMixin):
     def get_success_url(self):
         if self.success_url:
             url = self.success_url
@@ -84,17 +99,24 @@ class ModelFormSetMixin(FormSetMixin, MultipleObjectMixin):
     formfield_callback = None
 
     def get_context_data(self, **kwargs):
-        context = kwargs
+        context = {}
 
         if self.object_list:
             context['object_list'] = self.object_list
             context_object_name = self.get_context_object_name(self.get_queryset())
             if context_object_name:
                 context[context_object_name] = self.object_list
-        return context
+        context.update(kwargs)
 
-    def construct_formset(self):
-        return self.get_formset()(queryset=self.get_queryset(), **self.get_formset_kwargs())
+        # MultipleObjectMixin get_context_data() doesn't work when object_list
+        # is not provided in kwargs, so we skip MultipleObjectMixin and call
+        # ContextMixin directly.
+        return ContextMixin.get_context_data(self, **context)
+
+    def get_formset_kwargs(self):
+        kwargs = super(ModelFormSetMixin, self).get_formset_kwargs()
+        kwargs['queryset'] = self.get_queryset()
+        return kwargs
 
     def get_factory_kwargs(self):
         kwargs = super(ModelFormSetMixin, self).get_factory_kwargs()
@@ -128,18 +150,19 @@ class BaseInlineFormSetMixin(BaseFormSetMixin):
     can_delete = True
     save_as_new = False
 
+    def get_initial_data(self):
+        return {}
+
     def get_context_data(self, **kwargs):
-        context = kwargs
+        context = {}
 
         if self.object:
             context['object'] = self.object
             context_object_name = self.get_context_object_name(self.object)
             if context_object_name:
                 context[context_object_name] = self.object
-        return context
-
-    def construct_formset(self):
-        return self.get_formset()(instance=self.object, **self.get_formset_kwargs())
+        context.update(kwargs)
+        return super(BaseInlineFormSetMixin, self).get_context_data(**context)
 
     def get_inline_model(self):
         return self.inline_model
@@ -147,6 +170,8 @@ class BaseInlineFormSetMixin(BaseFormSetMixin):
     def get_formset_kwargs(self):
         kwargs = super(BaseInlineFormSetMixin, self).get_formset_kwargs()
         kwargs['save_as_new'] = self.save_as_new
+        kwargs['instance'] = self.object
+        del kwargs['initial']
         return kwargs
 
     def get_factory_kwargs(self):
